/**
 * BlackBerry Limited proprietary and confidential.
 * Do not reproduce without permission in writing.
 * Copyright (c) 2019 BlackBerry Ltd.
 * All rights reserved.
 */
package com.rim.platform.samples.rest.oidc;

import java.awt.Desktop;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLEncoder;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

import org.apache.commons.codec.binary.Base64;
import org.apache.http.HttpEntity;
import org.apache.http.HttpHeaders;
import org.apache.http.HttpStatus;
import org.apache.http.NameValuePair;
import org.apache.http.client.config.CookieSpecs;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.utils.URLEncodedUtils;
import org.apache.http.conn.ssl.NoopHostnameVerifier;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.oauth2.sdk.AccessTokenResponse;
import com.nimbusds.oauth2.sdk.AuthorizationCodeGrant;
import com.nimbusds.oauth2.sdk.ClientCredentialsGrant;
import com.nimbusds.oauth2.sdk.ErrorObject;
import com.nimbusds.oauth2.sdk.ParseException;
import com.nimbusds.oauth2.sdk.TokenErrorResponse;
import com.nimbusds.oauth2.sdk.TokenResponse;
import com.nimbusds.oauth2.sdk.auth.JWTAuthentication;
import com.nimbusds.oauth2.sdk.auth.PrivateKeyJWT;
import com.nimbusds.oauth2.sdk.id.ClientID;
import com.nimbusds.oauth2.sdk.token.BearerAccessToken;
import com.nimbusds.oauth2.sdk.token.Tokens;
import com.rim.platform.samples.exception.SampleException;

import net.minidev.json.JSONObject;

/**
 * Abstract base class for all REST samples.
 */
public abstract class AbstractRestSample
{
    /**
     * The EID endpoint discovery service
     */
    private OidcDiscoveryDocumentService oidcDiscoveryDocumentService = new OidcDiscoveryDocumentService();

    /**
     * A TrustManager that trusts all clients and servers. Use for testing only.
     */
    private TrustManager allTrustingTrustManager = new AllTrustingTrustManager();

    // Misc constants
    private static final String UTF_8 = "UTF-8";
    private static final String GRANT_TYPE = "grant_type";
    private static final String SCOPE = "scope";
    private static final String CLIENT_ASSERTION_TYPE = "client_assertion_type";
    private static final String CLIENT_ASSERTION = "client_assertion";
    private static final String REDIRECT_URI = "redirect_uri";
    private static final String CODE = "code";

    /**
     * Simple logger. See simplelogger.properties to change the logging settings
     */
    private static final Logger LOGGER = LoggerFactory.getLogger(AbstractRestSample.class);

    /**
     * Returns the {@link OidcDiscoveryDocumentService}
     * 
     * @return the {@link OidcDiscoveryDocumentService}
     */
    public OidcDiscoveryDocumentService getOidcDiscoveryDocumentService()
    {
        return oidcDiscoveryDocumentService;
    }

    /**
     * Returns the access token using the client secret
     * 
     * @param tokenEndpointUrl
     * @param clientId
     *            the client id
     * @param clientSecret
     *            the client secret set during client creation
     * @param scope
     *            the scope requested
     * @return the access token
     * @throws SampleException
     *             thrown if there is a problem getting the access token
     */
    protected String getAccessToken(String tokenEndpointUrl, String clientId, String clientSecret, String scope)
            throws SampleException
    {
        // Create the HTTP client and invoke the access token request
        try (CloseableHttpClient httpClient = createHttpClient())
        {
            // Create the POST request
            HttpPost post = new HttpPost(new URI(tokenEndpointUrl));
            post.setHeader(HttpHeaders.CONTENT_TYPE, URLEncodedUtils.CONTENT_TYPE);

            // URL encode the client secret
            String urlEncodedClientSecret = URLEncoder.encode(clientSecret, UTF_8);

            // Create the auth header with the clientId and the encoded secret
            String authHeader = Base64.encodeBase64String((clientId + ":" + urlEncodedClientSecret).getBytes());

            // Add the basic authorization header to the request
            post.setHeader(HttpHeaders.AUTHORIZATION, "Basic " + authHeader);

            // Create the post body for the request
            post.setEntity(getTokenUrlEncodedFormEntity(scope));

            // Execute request
            try (CloseableHttpResponse response = httpClient.execute(post))
            {
                // Get and check the status code
                int statusCode = response.getStatusLine().getStatusCode();
                HttpEntity responseEntity = response.getEntity();
                if (statusCode == HttpStatus.SC_OK)
                {
                    // Convert the response to a string
                    String responseBody = EntityUtils.toString(responseEntity);

                    // Process the response into the actual access code
                    return processAccessCodeResponse(responseBody);
                }
                else
                {
                    throw new SampleException("Failed to get access token from EID service, reason: " + statusCode + ", "
                            + (responseEntity != null ? EntityUtils.toString(responseEntity) : "No content"));
                }
            }
        }
        catch (UnsupportedEncodingException e)
        {
            throw new SampleException("Unable to create token authentication request", e);
        }
        catch (IOException e)
        {
            throw new SampleException("unable to obtain access token", e);
        }
        catch (URISyntaxException e)
        {
            throw new SampleException("Error processing " + tokenEndpointUrl, e);
        }
        catch (Exception e)
        {
            throw new SampleException("Unknown error while getting access code", e);
        }
    }

    /**
     * Returns the access token using a private key
     * 
     * @param tokenEndpointUrl
     * @param clientId
     *            the client id
     * @param privateKey
     * @param keyId
     * @param scope
     *            the scope requested
     * @return the access token
     * @throws SampleException
     *             thrown if there is a problem getting the access token
     */
    protected String getAccessToken(String tokenEndpointUrl, String clientId, PrivateKey privateKey, String keyId, String scope)
            throws SampleException
    {
        // Create the HTTP client and invoke the access token request
        try (CloseableHttpClient httpClient = createHttpClient())
        {
            // Create the POST request
            HttpPost post = new HttpPost(new URI(tokenEndpointUrl));
            post.setHeader(HttpHeaders.CONTENT_TYPE, URLEncodedUtils.CONTENT_TYPE);

            // Create the private key JWT object
            PrivateKeyJWT pkJwt = new PrivateKeyJWT(new ClientID(clientId), new URI(tokenEndpointUrl), JWSAlgorithm.RS256,
                    (RSAPrivateKey) privateKey, keyId, null);

            // Create the post body for the request
            post.setEntity(getTokenUrlEncodedFormEntity(pkJwt, scope));

            // Execute the request
            try (CloseableHttpResponse response = httpClient.execute(post))
            {
                // Get and check the status code
                int statusCode = response.getStatusLine().getStatusCode();
                HttpEntity responseEntity = response.getEntity();
                if (statusCode == HttpStatus.SC_OK)
                {
                    // Convert the response to a string
                    String responseBody = EntityUtils.toString(responseEntity);

                    // Process the response into the actual access code
                    return processAccessCodeResponse(responseBody);
                }
                else
                {
                    throw new SampleException("Failed to get access token from EID service, reason: " + statusCode + ", "
                            + (responseEntity != null ? EntityUtils.toString(responseEntity) : "No content"));
                }
            }
        }
        catch (UnsupportedEncodingException e)
        {
            throw new SampleException("Unable to create token authentication request", e);
        }
        catch (IOException e)
        {
            throw new SampleException("Unable to obtain access token", e);
        }
        catch (URISyntaxException e)
        {
            throw new SampleException("Error processing " + tokenEndpointUrl, e);
        }
        catch (Exception e)
        {
            throw new SampleException("Unknown error while getting access code", e);
        }
    }

    /**
     * Returns the access token using a client secret and authorization code
     * 
     * @param tokenEndpointUrl
     * @param clientId
     * @param clientSecret
     * @param redirectUri
     * @param code
     * @return
     * @throws SampleException
     */
    public String getAccessToken(String tokenEndpointUrl, String clientId, String clientSecret, String redirectUri, String code)
            throws SampleException
    {
        // Create the HTTP client and invoke the access token request
        try (CloseableHttpClient httpClient = createHttpClient())
        {
            // Create the POST request
            HttpPost post = new HttpPost(new URI(tokenEndpointUrl));
            post.setHeader(HttpHeaders.CONTENT_TYPE, URLEncodedUtils.CONTENT_TYPE);

            // URL encode the client secret
            String urlEncodedClientSecret = URLEncoder.encode(clientSecret, UTF_8);

            // Create the auth header with the clientId and the encoded secret
            String authHeader = Base64.encodeBase64String((clientId + ":" + urlEncodedClientSecret).getBytes());

            // Add the basic authorization header to the request
            post.setHeader(HttpHeaders.AUTHORIZATION, "Basic " + authHeader);

            // Create the post body for the request
            post.setEntity(getTokenUrlEncodedFormEntity(code, redirectUri));

            // Execute request
            try (CloseableHttpResponse response = httpClient.execute(post))
            {
                // Get and check the status code
                int statusCode = response.getStatusLine().getStatusCode();
                HttpEntity responseEntity = response.getEntity();
                if (statusCode == HttpStatus.SC_OK)
                {
                    // Convert the response to a string
                    String responseBody = EntityUtils.toString(responseEntity);

                    // Process the response into the actual access code
                    return processAccessCodeResponse(responseBody);
                }
                else
                {
                    throw new SampleException("Failed to get access token from EID service, reason: " + statusCode + ", "
                            + (responseEntity != null ? EntityUtils.toString(responseEntity) : "No content"));
                }
            }
        }
        catch (UnsupportedEncodingException e)
        {
            throw new SampleException("Unable to create token authentication request", e);
        }
        catch (IOException e)
        {
            throw new SampleException("unable to obtain access token", e);
        }
        catch (URISyntaxException e)
        {
            throw new SampleException("Error processing " + tokenEndpointUrl, e);
        }
        catch (Exception e)
        {
            throw new SampleException("Unknown error while getting access code", e);
        }
    }

    /**
     * Login to the Idp auth endpoint
     * 
     * @param eidUrl
     * @param tenantId
     * @param clientId
     * @param redirectUri
     * @param scope
     * @throws IOException
     * @throws SampleException
     */
    public void login(String eidUrl, String tenantId, String clientId, String redirectUri, String scope) throws SampleException
    {
        try
        {
            String urlTemplate = "%s/op/tenant/%s/auth?client_id=%s&redirect_uri=%s&scope=%s&prompt=consent&response_type=code";
            String url = String.format(urlTemplate, eidUrl, tenantId, clientId, URLEncoder.encode(redirectUri, UTF_8),
                    URLEncoder.encode(scope, UTF_8));

            // Needs Gnome libraries for running on Linux
            if (!Desktop.isDesktopSupported())
            {
                LOGGER.error("Cannot open browser window");
                throw new SampleException("Cannot open browser window");
            }

            Desktop desktop = Desktop.getDesktop();
            desktop.browse(new URI(url));
        }
        catch (IOException | URISyntaxException e)
        {
            LOGGER.error("Exception caught: {}", e.getMessage());
            throw new SampleException(e.getMessage());
        }
    }

    /**
     * Invokes the passed URL with the passed access token. The URL is processed as a GET request.
     * 
     * @param url
     *            the URL to invoke
     * @param accessToken
     *            the access token
     * @return the response from the server
     * @throws IOException
     * @throws NoSuchAlgorithmException
     * @throws NoSuchProviderException
     * @throws KeyManagementException
     */
    protected CloseableHttpResponse invokeRestApi(String url, String accessToken)
            throws IOException, NoSuchAlgorithmException, NoSuchProviderException, KeyManagementException
    {
        // Initialize the SSLContext and SSLConnectionSocketFactory to trust all
        SSLContext sslContext = SSLContext.getInstance("TLSv1.2", "SunJSSE");
        sslContext.init(null, new TrustManager[] { allTrustingTrustManager }, null);
        SSLConnectionSocketFactory sslConnectionSocketFactory =
                new SSLConnectionSocketFactory(sslContext, NoopHostnameVerifier.INSTANCE);

        // Create the HTTP client
        CloseableHttpClient httpClient = createHttpClient(sslConnectionSocketFactory);

        // Create the request
        HttpGet request = new HttpGet(url);
        request.addHeader("Authorization", "Bearer " + accessToken);

        // Invoke the request
        LOGGER.debug("Sending request to {}", url);
        return httpClient.execute(request);
    }

    /**
     * Processes the response body from an access code request and returns the actual access code
     * 
     * @param responseBody
     *            the response body from an access code request
     * @return the actual access code
     * @throws SampleException
     */
    private String processAccessCodeResponse(String responseBody) throws SampleException
    {
        try
        {
            // Read the JSON response body into a Map
            ObjectMapper mapper = new ObjectMapper();
            Map<String, Object> resMap = mapper.readValue(responseBody, new TypeReference<Map<String, Object>>()
            {
            });

            // Parse the Map into a TokenResponse
            TokenResponse tokenResponse = TokenResponse.parse(new JSONObject(resMap));

            // Check if an error response
            if (tokenResponse instanceof TokenErrorResponse)
            {
                ErrorObject error = ((TokenErrorResponse) tokenResponse).getErrorObject();
                throw new SampleException("failed to get the Access token, error code [" + error.getCode()
                        + "], error description [" + error.getDescription() + "], http code [" + error.getHTTPStatusCode() + "]");
            }

            // Check is some other type of response
            if (!(tokenResponse instanceof AccessTokenResponse))
            {
                throw new SampleException("tokenResponse " + tokenResponse + " is not an instance of AccessTokenResponse");
            }

            // Get the tokens from the response
            Tokens tokens = ((AccessTokenResponse) tokenResponse).getTokens();
            if (tokens == null)
            {
                throw new SampleException("tokens is null");
            }

            // Get the Bearer token
            BearerAccessToken bearerAccessToken = tokens.getBearerAccessToken();
            if (bearerAccessToken == null)
            {
                throw new SampleException("bearerAccessToken is null");
            }

            // Return the value of the Bearer token
            return bearerAccessToken.getValue();
        }
        catch (ParseException e)
        {
            throw new SampleException("Failed to parse the response [" + responseBody + "] into json", e);
        }
        catch (IOException e)
        {
            throw new SampleException("Failed to map token response [" + responseBody + "]", e);
        }
    }

    /**
     * Creates URL encoded form entity for the passed scope
     * 
     * @param scope
     *            the scope
     * @return the URL encoded form entity
     * @throws UnsupportedEncodingException
     */
    private UrlEncodedFormEntity getTokenUrlEncodedFormEntity(String scope) throws UnsupportedEncodingException
    {
        List<NameValuePair> nameValuePairs = new ArrayList<>();

        nameValuePairs.add(new BasicNameValuePair(GRANT_TYPE, ClientCredentialsGrant.GRANT_TYPE.toString()));
        nameValuePairs.add(new BasicNameValuePair(SCOPE, scope));

        return new UrlEncodedFormEntity(nameValuePairs, UTF_8);
    }

    /**
     * Creates URL encoded form entity for the passed PrivateKeyJWT and scope
     * 
     * @param privateKeyJWT
     *            the PrivateKeyJWT
     * @param scope
     *            the scope
     * @return the URL encoded form entity
     * @throws UnsupportedEncodingException
     */
    private UrlEncodedFormEntity getTokenUrlEncodedFormEntity(PrivateKeyJWT privateKeyJWT, String scope)
            throws UnsupportedEncodingException
    {
        List<NameValuePair> nvps = new ArrayList<>();

        nvps.add(new BasicNameValuePair(GRANT_TYPE, ClientCredentialsGrant.GRANT_TYPE.toString()));
        nvps.add(new BasicNameValuePair(CLIENT_ASSERTION, privateKeyJWT.getClientAssertion().serialize()));
        nvps.add(new BasicNameValuePair(CLIENT_ASSERTION_TYPE, JWTAuthentication.CLIENT_ASSERTION_TYPE));
        nvps.add(new BasicNameValuePair(SCOPE, scope));

        return new UrlEncodedFormEntity(nvps, UTF_8);
    }

    /**
     * Creates URL encoded form entity for the passed code and redirect uri
     * 
     * @param code
     * @param redirectUri
     * @return
     * @throws UnsupportedEncodingException
     */
    private UrlEncodedFormEntity getTokenUrlEncodedFormEntity(String code, String redirectUri) throws UnsupportedEncodingException
    {
        List<NameValuePair> nvps = new ArrayList<>();

        nvps.add(new BasicNameValuePair(GRANT_TYPE, AuthorizationCodeGrant.GRANT_TYPE.toString()));
        nvps.add(new BasicNameValuePair(CODE, code));
        nvps.add(new BasicNameValuePair(REDIRECT_URI, redirectUri));

        return new UrlEncodedFormEntity(nvps, UTF_8);
    }

    /**
     * Creates a basic HTTP client
     * 
     * @return the HTTP client
     */
    private CloseableHttpClient createHttpClient()
    {
        return createHttpClient(null);
    }

    /**
     * Creates a basic HTTP client with the passed SSLConnectionSocketFactory
     * 
     * @param sslConnectionSocketFactory
     *            the SSLConnectionSocketFactory to use, null if a SSLConnectionSocketFactory is not required
     * @return the HTTP client
     */
    private CloseableHttpClient createHttpClient(SSLConnectionSocketFactory sslConnectionSocketFactory)
    {
        return HttpClients.custom().setDefaultRequestConfig(RequestConfig.custom().setCookieSpec(CookieSpecs.STANDARD).build())
                .setSSLSocketFactory(sslConnectionSocketFactory).build();
    }

    /**
     * Simple class to trust all clients and servers for testing purposes
     */
    private class AllTrustingTrustManager implements X509TrustManager
    {
        @Override
        public void checkClientTrusted(X509Certificate[] chain, String authType) throws CertificateException
        {
            // All trusted
        }

        @Override
        public void checkServerTrusted(X509Certificate[] chain, String authType) throws CertificateException
        {
            // All trusted
        }

        @Override
        public X509Certificate[] getAcceptedIssuers()
        {
            return new X509Certificate[0];
        }
    }
}
