/**
 * BlackBerry Limited proprietary and confidential.
 * Do not reproduce without permission in writing.
 * Copyright (c) 2019 BlackBerry Ltd.
 * All rights reserved.
 */
package com.rim.platform.samples.rest.oidc;

import static com.github.tomakehurst.wiremock.client.WireMock.aResponse;
import static com.github.tomakehurst.wiremock.client.WireMock.get;
import static com.github.tomakehurst.wiremock.client.WireMock.urlMatching;
import static com.github.tomakehurst.wiremock.core.WireMockConfiguration.wireMockConfig;

import java.net.URL;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpResponse;
import org.apache.http.util.EntityUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.tomakehurst.wiremock.WireMockServer;
import com.github.tomakehurst.wiremock.client.ResponseDefinitionBuilder;
import com.github.tomakehurst.wiremock.common.FileSource;
import com.github.tomakehurst.wiremock.core.WireMockConfiguration;
import com.github.tomakehurst.wiremock.extension.Parameters;
import com.github.tomakehurst.wiremock.extension.ResponseDefinitionTransformer;
import com.github.tomakehurst.wiremock.http.QueryParameter;
import com.github.tomakehurst.wiremock.http.Request;
import com.github.tomakehurst.wiremock.http.ResponseDefinition;

public class SampleWithAuthorizationCodeAndClientSecret extends AbstractRestSample
{
    private static Logger LOGGER = LoggerFactory.getLogger(SampleWithClientCredsAndClientSecret.class);

    private WireMockServer mockCallbackListener;

    public void startMockCallbackListener(String tokenEndpointUrl, String clientId, String clientSecret, String callbackUrl,
            String restApiUrl) throws Exception
    {
        URL url = new URL(callbackUrl);
        int callbackPort = url.getPort() == -1 ? 443 : url.getPort();

        // Assumes localhost for callback host
        final WireMockConfiguration config = wireMockConfig().httpsPort(callbackPort)
                .extensions(new CallbackTransformer(tokenEndpointUrl, clientId, clientSecret, callbackUrl, restApiUrl));
        mockCallbackListener = new WireMockServer(config);

        mockCallbackListener.stubFor(get(urlMatching(url.getPath() + ".*"))
                .willReturn(aResponse().withStatus(200).withTransformers("callbackTransformer")));
        // Need to send response for /favicon since browser expects it
        mockCallbackListener.stubFor(get(urlMatching("/favicon.ico")).willReturn(aResponse().withStatus(200)));

        mockCallbackListener.start();
    }

    public void stopMockCallbackListener()
    {
        if (mockCallbackListener != null)
        {
            mockCallbackListener.stop();
            Runtime.getRuntime().exit(0);
        }
    }

    public class CallbackTransformer extends ResponseDefinitionTransformer
    {
        private static final String TOKEN = "token";
        private static final String LOGOUT_URI = "session/end";
        private String tokenEndpointUrl;
        private String clientId;
        private String clientSecret;
        private String callbackUrl;
        private String restApiUrl;

        public CallbackTransformer(String tokenEndpointUrl, String clientId, String clientSecret, String callbackUrl,
                String restApiUrl)
        {
            this.tokenEndpointUrl = tokenEndpointUrl;
            this.clientId = clientId;
            this.clientSecret = clientSecret;
            this.callbackUrl = callbackUrl;
            this.restApiUrl = restApiUrl;
        }

        @Override
        public ResponseDefinition transform(Request request, ResponseDefinition responseDefinition, FileSource files,
                Parameters parameters)
        {
            try
            {
                QueryParameter codeParam = request.queryParameter("code");
                String code = String.format(codeParam.firstValue());

                // Get the access token using the code from the redirect
                String accessToken = getAccessToken(tokenEndpointUrl, clientId, clientSecret, callbackUrl, code);

                // Invoke the rest api
                HttpResponse response = invokeRestApi(restApiUrl, accessToken);
                String result = EntityUtils.toString(response.getEntity());

                // Schedule a task to close the mock server and stop the application once we get back a response
                // from the rest api
                ScheduledExecutorService executor = Executors.newSingleThreadScheduledExecutor();
                executor.schedule(() -> {
                    stopMockCallbackListener();
                }, 1, TimeUnit.MINUTES);

                return new ResponseDefinitionBuilder().withStatus(200).withBody(displayRestApiResponse(tokenEndpointUrl, result))
                        .build();
            }
            catch (Exception e)
            {
                LOGGER.error("Exception caught: {}", e);
                return null;
            }
        }

        private String displayRestApiResponse(final String tokenEndpointUrl, final String result)
        {
            String logoutUrl = StringUtils.removeEndIgnoreCase(tokenEndpointUrl, TOKEN) + LOGOUT_URI;
            return "<html><body>" + "<p>" + result + "</p>" + "<a href=\"" + logoutUrl + "\">Logout</a>" + "</body></html>";
        }

        @Override
        public boolean applyGlobally()
        {
            return false;
        }

        @Override
        public String getName()
        {
            return "callbackTransformer";
        }
    }

    public static void main(String[] args) throws Exception
    {
        // Create the sample application
        SampleWithAuthorizationCodeAndClientSecret sample = new SampleWithAuthorizationCodeAndClientSecret();

        // Create CLI options
        Options options = new Options();
        options.addOption(Option.builder("t").desc("Tenant id").hasArg().required().build());
        options.addOption(Option.builder("c").desc("Client id").hasArg().required().build());
        options.addOption(Option.builder("cs").desc("Client secret").hasArg().required().build());
        options.addOption(Option.builder("eid").desc("EID endpoint - Optional").hasArg().build());
        options.addOption(Option.builder("uem").desc("UEM endpoint").hasArg().required().build());
        options.addOption(Option.builder("cb").desc("Callback url").hasArg().required().build());

        try
        {
            // Parse the command line
            CommandLineParser parser = new DefaultParser();
            CommandLine cmd = parser.parse(options, args);

            // Get the values from the command line
            String eidUrl = cmd.getOptionValue("eid", "https://idp.blackberry.com");
            String tenantId = cmd.getOptionValue("t");
            String clientId = cmd.getOptionValue("c");
            String clientSecret = cmd.getOptionValue("cs");
            String restApiUrl = cmd.getOptionValue("uem") + "/" + tenantId + "/api/v1/ping";
            String callbackUrl = cmd.getOptionValue("cb");
            String scope = "openid MDMBWS.All";

            LOGGER.info("EID URL: {}", eidUrl);
            LOGGER.info("Tenant Id: {}", tenantId);
            LOGGER.info("Client Id: {}", clientId);
            LOGGER.info("Client Secret: {}", clientSecret);
            LOGGER.info("REST API URL: {}", restApiUrl);
            LOGGER.info("Callback url: {}", callbackUrl);
            LOGGER.info("Scope: {}", scope);

            // Get the token endpoint URL for the tenant from the discovery document
            LOGGER.info("Getting the token endpoint URL for the tenant from the discovery document");
            String endpointUrl = sample.getOidcDiscoveryDocumentService().getEndpointUrl(eidUrl,
                    OidcDiscoveryDocumentService.EndpointKey.TOKEN_ENDPOINT, tenantId);

            // Start the mock callback server
            sample.startMockCallbackListener(endpointUrl, clientId, clientSecret, callbackUrl, restApiUrl);

            // Open the login page
            sample.login(eidUrl, tenantId, clientId, callbackUrl, scope);
        }
        catch (ParseException e)
        {
            LOGGER.error(e.getMessage(), e);
            HelpFormatter formatter = new HelpFormatter();
            formatter.printHelp("sample", options);
        }
    }
}
