﻿param([switch]$force=$false, [switch]$debug=$false)

$tenantGuid 		= "Set-Tenant-Guid-Here"
$eidBaseUrl 		= "https://idp.blackberry.com"
$clientId 			= "Set-Client-Id-Here";
$clientSecret 		= "Set-Client-Secret-Here";
$powerAppURL 		= "https://localhost:9443/cb"
$tokenEndpoint 		= $eidBaseUrl + "/op/tenant/{0}/token" -f $tenantGuid
$authEndpoint 		= $eidBaseUrl + "/op/tenant/{0}/auth" -f $tenantGuid
# We need to include openid and offline_access in the scopes for authorization_code
$scope				= "openid MDMBWS.All offline_access"
# UEM On-premise API port is 18084 and Cloud API port is 443
$restApiUrl 		= "https://Set-Host-Here:18084/{0}/api/v1/ping" -f $tenantGuid;

Function Refresh-BlackBerryTokens{
[CmdletBinding()]
param($clientID,$powerAppURL,$clientSecret,$refreshToken)

    $encodedSecret = [uri]::EscapeDataString($clientSecret)
    $basicHeader = @{"Authorization" = ("Basic",[Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes(($clientId, $encodedSecret -join ":"))) -join " ")}
    $tokenRequest = @{redirect_uri=$powerAppURL; grant_type="refresh_token"; refresh_token=$refreshToken}

    Write-Debug "Attempting to get an access token with refresh_token = $refreshToken"

    try { 
		[Net.ServicePointManager]::SecurityProtocol = [Net.SecurityProtocolType]::Ssl3, [Net.SecurityProtocolType]::Tls, [Net.SecurityProtocolType]::Tls11, [Net.SecurityProtocolType]::Tls12
        $result = Invoke-RestMethod -Uri $tokenEndpoint -Method Post -Body $tokenRequest -ContentType "application/x-www-form-urlencoded" -Headers $basicHeader -ErrorAction STOP
    }
    catch{
        Write-Warning "Something didn't work"
        Write-debug "Test the -body params for the Rest command"
    }
    
    if ($result.access_token -and $result.refresh_token){
        $global:accessToken = $result.access_token
        $global:refreshToken = $result.refresh_token
        Write-Output "Response from the token endpoint:"
        $result
    }
}

Function Get-BlackBerryTokens{
[CmdletBinding()]
param($clientID,$powerAppURL,$clientSecret,$authCode)

    $encodedSecret = [uri]::EscapeDataString($clientSecret) 
    $basicHeader = @{"Authorization" = ("Basic",[Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes(($clientId, $encodedSecret -join ":"))) -join " ")}
    $tokenRequest = @{redirect_uri=$powerAppURL; grant_type="authorization_code"; code=$authCode}

    Write-Debug "Attempting to get an access token with authorization code = $authCode"

    try { 
		[Net.ServicePointManager]::SecurityProtocol = [Net.SecurityProtocolType]::Ssl3, [Net.SecurityProtocolType]::Tls, [Net.SecurityProtocolType]::Tls11, [Net.SecurityProtocolType]::Tls12
        $result = Invoke-RestMethod -Uri $tokenEndpoint -Method Post -Body $tokenRequest -ContentType "application/x-www-form-urlencoded" -Headers $basicHeader -ErrorAction STOP
    }
    catch{
        Write-Warning "Something didn't work"
        Write-debug "Test the -body params for the Rest command"
    }
    
    if ($result.access_token -and $result.refresh_token){
        $global:accessToken = $result.access_token
        $global:refreshToken = $result.refresh_token
        Write-Output "Response from the token endpoint:"
        $result
    }
}

Function Get-BlackBerryAuthorization{
param($ClientID,$powerAppURL)
    
    If($ClientID -eq $null){
        Write-Warning "Must provide the `-ClientID of your BlackBerry App to display a login Window"
        break}

#   Show OAuth Login Window

    Add-Type -AssemblyName System.Windows.Forms
    $url = $authEndpoint + "?client_id=$clientID&redirect_uri=$powerAppURL&scope=$scope&prompt=consent&response_type=code"
    $form = New-Object -TypeName System.Windows.Forms.Form -Property @{Width=440;Height=440}
    $web  = New-Object -TypeName System.Windows.Forms.WebBrowser -Property @{Width=420;Height=400;Url=($url -f ($Scope -join "%20")) }
    $DocComp  = {
       $Global:uri = $web.Url.AbsoluteUri
       if ($Global:Uri -match "error=[^&]*|code=[^&]*") {$form.Close()}
    }
    $web.ScriptErrorsSuppressed = $true
    $web.Add_DocumentCompleted($DocComp)
    $form.Controls.Add($web)
    $form.Add_Shown({$form.Activate()})
    $form.ShowDialog() | Out-Null

    #After this, there should be a variable called $uri, which has the authorization code!

    #(?<=code=)(.*)(?=&) Note: This is the original regex which didn't work correctly
    $regex = '(?<=code=)([^&]+)'
    if (-not ($uri -match $regex)) {
        Write-Debug "Did not get an authcode"
        exit
    }
    $authCode  = $Matches[0]
    $global:authCode = $authCode
}

Function Connect-BlackBerryAccount {
<#
.Synopsis
    Use this cmdlet to connect to BlackBerry Identity to obtain access tokens to access BlackBerry APIs via PowerShell
.DESCRIPTION
 With one cmdlet, you can connect to BlackBerry  to access BlackBerry REST APIs. 
.EXAMPLE
    Connect-BlackBerryAccount -ClientID [String] -PowerAppURL [(Redirect URL)] -ClientSecret [ClientSecret]
    Sign up at blackberry.com and make a new client ID, which is needed to access BlackBerry APIs.  While you're there, specify a redirect URL, which should be the URL of your powershell App. You'll receive a ClientID, ClientSecret and RedirectURL, which you must provide to this cmdlet.
    Upon running, a Internet Explorer com object window will be displayed, prompting you to login and authorize your Client ID (PowerShell, effectively) to interact with your BlackBerry account.
    Behind the scenes, this cmdlet will retrieve an authorization code, convert it to an access token and a refresh token. The refresh token is stored safely within your profile.  BlackBerry API calls require the access token and it will be automatically refreshed when needed.
.EXAMPLE
    Connect-BlackBerryAccount -Force
    If you need to renew your refresh token, then rerun the cmdlet with -Force

    Run this command in your powershell before running this script
    set-executionpolicy remotesigned
#>
[CmdletBinding()]
param($clientID,$powerAppURL,$clientSecret,[Switch]$force)

    $configDir = "$Env:AppData\WindowsPowerShell\Modules\BlackBerryAccount\0.1\Config.ps1xml"
    if (-not (Test-Path $configDir) -or $force){
        if ($force){"Force detected"}
        New-item -Force -Path "$configDir" -ItemType File

        Write-Debug "Authenticating User"
        Get-BlackBerryAuthorization -ClientID $ClientID -powerAppURL $powerAppURL

        Get-BlackBerryTokens -ClientID $ClientID -clientSecret $clientSecret -powerAppURL $powerAppURL -authCode $authCode

        #store the refresh token
        $password = ConvertTo-SecureString $refreshToken -AsPlainText -Force
        $password | ConvertFrom-SecureString | Export-Clixml $configDir -Force
        Write-Debug "Stored an encrypted refresh token"
    }
    else{
        Write-Debug "This app is already connected to BlackBerry"
        try {
             $password = Import-Clixml -Path $configDir -ErrorAction STOP | ConvertTo-SecureString
        }
        catch {
            Write-Warning "Corrupt Password file found, rerun with -Force to fix this"
            BREAK
       }
       $Ptr = [System.Runtime.InteropServices.Marshal]::SecureStringToCoTaskMemUnicode($password)
       $result = [System.Runtime.InteropServices.Marshal]::PtrToStringUni($Ptr)
       [System.Runtime.InteropServices.Marshal]::ZeroFreeCoTaskMemUnicode($Ptr)
       $global:refreshToken = $result 
       Refresh-BlackBerryTokens -clientID $clientID -clientSecret $clientSecret -powerAppURL $powerAppURL -refreshToken $result

       #store the refresh token since it is rotated
       $password = ConvertTo-SecureString $refreshToken -AsPlainText -Force
       $password | ConvertFrom-SecureString | Export-Clixml $configDir -Force
       Write-Debug "Stored an encrypted refresh token"
   }
}

# First connect to BlackBerry. This makes the initial connection or refreshes tokens.
Connect-BlackBerryAccount -clientID $clientID -clientSecret $clientSecret -powerAppURL $powerAppURL -force:$force -debug:$debug

# Invoke the REST Api with the access token in the Authorization header
try
{
	# Add Trust All
	add-type @"
	using System.Net;
	using System.Security.Cryptography.X509Certificates;
	public class TrustAllCertsPolicy : ICertificatePolicy {
		public bool CheckValidationResult(
			ServicePoint srvPoint, X509Certificate certificate,
			WebRequest request, int certificateProblem) {
			return true;
		}
	}
"@
	[System.Net.ServicePointManager]::CertificatePolicy = New-Object TrustAllCertsPolicy
	[Net.ServicePointManager]::SecurityProtocol = [Net.SecurityProtocolType]::Ssl3, [Net.SecurityProtocolType]::Tls, [Net.SecurityProtocolType]::Tls11, [Net.SecurityProtocolType]::Tls12

	$authHeader = @{"Authorization" = ("Bearer", $accessToken -join " ")}

	Write-Output ""
	Write-Output "Attempting to access Public REST API"
	Write-Output "Request: $restApiUrl"

	$response = Invoke-RestMethod -Method GET -Uri $restApiUrl -Headers $authHeader -ErrorAction STOP

	Write-Output ""
	Write-Output "Received response for rest api: $restApiUrl"
	Write-Output $response
	Write-Output ""
}
catch
{
	Write-Error ("The command could not be completed.  `r`nReason: " + $error.exception.message + " `r`nDetails: "+ $error.errordetails);
	exit(1);
}

Write-Output ""
Write-Output("Complete!");
